@@ -244,7 +244,7 @@ def StoreCacheDatum(ticker, objectName, datum, expiry=None, metadata=None):
 	fp = GetFilepath(ticker, objectName, object=datum, prune=True)
 
 	if verbose:
-		print("- storing {} as {} at {}".format(objectName, ext, fp))
+		print("- storing {} at {}".format(objectName, fp))
 
 	d = _ReadData(ticker, objectName)
 	if d is None:
@@ -298,27 +298,36 @@ def StoreCachePackedDatum(ticker, objectName, datum, expiry=None, metadata=None)
 	# Read cached metadata
 	fp = GetFilepath(ticker, objectName)
 	pkData = _ReadPackedData(ticker, objectName)
-	if (not pkData is None) and (objectName in pkData):
+	if pkData is None:
+		pkData = {}
+		objData = None
+		data = None
+	elif objectName in pkData:
 		objData = pkData[objectName]
 		data   = objData["data"]
-		md     = objData["metadata"] if "metadata" in objData else None
+		if metadata is None:
+			metadata = objData["metadata"] if "metadata" in objData else None
 		if expiry is None:
 			expiry = objData["expiry"] if "expiry" in objData else None
 	else:
-		pkData = {}
+		objData = None
+		data = None
 		md = None
 
-	if metadata is None:
-		# Persist the cached metadata
-		metadata = md
+	if objData is None:
+		objData = {"data":datum}
+		if not metadata is None:
+			objData["metadata"] = metadata
+		if not expiry is None:
+			objData["expiry"] = expiry
+	else:
+		objData["data"] = datum
+		if not metadata is None:
+			objData["metadata"] = metadata
+		if not expiry is None:
+			objData["expiry"] = expiry
 
-	# Write
-	d = {"data":datum}
-	if not metadata is None:
-		d["metadata"] = metadata
-	if not expiry is None:
-		d["expiry"] = expiry
-	pkData[objectName] = d
+	pkData[objectName] = objData
 	with open(fp, 'wb') as outData:
 		pickle.dump(pkData, outData, 4)
 
@@ -332,7 +341,11 @@ def ReadCacheMetadata(ticker, objectName, key):
 			data   = objData["data"]
 			md     = objData["metadata"] if "metadata" in objData else None
 	else:
-		data,md = _ReadData(ticker, objectName, return_metadata_too=True)
+		d = _ReadData(ticker, objectName)
+		md = d["metadata"]
+		if verbose:
+			print("ReadCacheMetadata() read md as:")
+			print(md)
 	if md is None:
 		return None
 	elif not key in md:
@@ -354,6 +367,10 @@ def WriteCacheMetadata(ticker, objectName, key, value):
 	else:
 		d["metadata"] = {key:value}
 
+	if verbose:
+		print("WriteCacheMetadata() updated md to:")
+		print(d["metadata"])
+
 	fp = GetFilepath(ticker, objectName)
 	if fp.endswith(".json"):
 		with open(fp, 'w') as outData:
